@@ -184,14 +184,22 @@ InitialiseRFBConnection(int sock)
 	    return False;
 	  }
 	} else {
+	  if(isatty(0))
+	  {
 	    passwd = getpass("Password: ");
-	    if (strlen(passwd) == 0) {
-		fprintf(stderr,"%s: Reading password failed\n",programName);
-		return False;
-	    }
-	    if (strlen(passwd) > 8) {
-		passwd[8] = '\0';
-	    }
+	  }else{
+	    static char password[128];
+	    passwd = fgets(password, sizeof(password)-1, stdin);
+	  }
+
+	  if (strlen(passwd) == 0) {
+	    fprintf(stderr,"%s: Reading password failed\n",programName);
+	    return False;
+	  }
+
+	  if (strlen(passwd) > 8) {
+	    passwd[8] = '\0';
+	  }
 	}
 
 	vncEncryptBytes(challenge, passwd);
@@ -341,6 +349,8 @@ SendPointerEvent(int x, int y, int buttonMask)
 {
     rfbPointerEventMsg pe;
 
+    last_event_time=time(0);
+
     pe.type = rfbPointerEvent;
     pe.buttonMask = buttonMask;
 #if 0
@@ -349,6 +359,7 @@ SendPointerEvent(int x, int y, int buttonMask)
 #endif
     pe.x = Swap16IfLE(x);
     pe.y = Swap16IfLE(y);
+
     return WriteExact(rfbsock, (char *)&pe, sz_rfbPointerEventMsg);
 }
 
@@ -362,6 +373,8 @@ SendKeyEvent(CARD32 key, Bool down)
 {
     rfbKeyEventMsg ke;
 
+    last_event_time=time(0);
+
     ke.type = rfbKeyEvent;
     ke.down = down ? 1 : 0;
     ke.key = Swap32IfLE(key);
@@ -413,6 +426,8 @@ SendClientCutText(char *str, int len)
 {
     rfbClientCutTextMsg cct;
 
+    last_event_time=time(0);
+
     if (trimsel && LooksLikeURL(str, len)) {
 	/*
 	 * Trim leading and trailing spaces.
