@@ -32,6 +32,21 @@
 #include <X11/cursorfont.h>
 #include <X11/keysym.h>
 
+#ifdef HAVE_XIDLE
+#include "xidle.h"
+
+int server_has_xidle=0;
+#endif
+
+#ifdef HAVE_MIT_SCREENSAVER
+#include <X11/extensions/scrnsaver.h>
+
+int server_has_mitscreensaver=0;
+#endif
+
+#ifdef HAVE_XINERAMA
+#include <X11/extensions/Xinerama.h>
+#endif
 
 Display *dpy;
 static Window topLevel;
@@ -43,9 +58,13 @@ static Bool modifierPressed[256];
 static Bool HandleTopLevelEvent(XEvent *ev);
 static Bool HandleRootEvent(XEvent *ev);
 static int displayWidth, displayHeight;
+static int x_offset=0, y_offset=0;
 static int grabbed;
 Cursor  grabCursor;
 
+#define XROOT(E) ((E).x_root - x_offset)
+#define YROOT(E) ((E).y_root - y_offset)
+
 enum edge_enum edge = EDGE_EAST;
 int edge_width=1;
 int restingx=-1;
@@ -56,6 +75,8 @@ int scroll_lines=1;
 int mac_mode=0;
 int hidden=0;
 
+int last_event_time = 0;
+
 KeySym grabkeysym=XK_F12;
 KeyCode grabkey;
 int grabmod = ControlMask;
@@ -82,8 +103,19 @@ int requested_desktop = -2;
 int current_desktop = -2;
 int current_number_of_desktops = -2;
 
+int remote_is_locked = 0;
+
 static void ungrabit(int x, int y, Window warpWindow);
 
+typedef int(*xerrorhandler)(Display *, XErrorEvent *);
+
+int warn_about_hotkey(Display *dpy, XErrorEvent *ev)
+{
+  grabkeysym=0;
+  fprintf(stderr,"Warning: Failed to bind x2vnc hotkey, hotkey disabled.\n");
+  return 0;
+}
+
 /*
  * This variable is true (1) if the mouse is on the same screen as the one
  * we're monitoring, or if there is only one screen on the X server.
@@ -112,6 +144,37 @@ Bool CreateXWindow(void)
 	    programName, XDisplayName(displayname));
     return False;
   }
+
+  /*
+   * check extensions
+   */
+#ifdef HAVE_XIDLE
+ {
+   int x,y;
+   server_has_xidle=XidleQueryExtension(dpy, &x,&y);
+ }
+#endif
+
+#ifdef HAVE_MIT_SCREENSAVER
+ {
+   int x,y;
+   server_has_mitscreensaver=XScreenSaverQueryExtension(dpy, &x,&y);
+#ifdef DEBUG
+   fprintf(stderr,"MIT-SCREEN-SAVER = %d\n",server_has_mitscreensaver);
+#endif
+ }
+#endif
+
+  if(noblank
+#ifdef HAVE_XIDLE
+     && (!server_has_xidle)
+#endif
+#ifdef HAVE_MIT_SCREENSAVER
+     && (!server_has_mitscreensaver)
+#endif
+    )
+    fprintf(stderr,"-noblank option used, but no approperiate extensions found:\n"
+	    " x2vnc will keep the remote screen active at all times.\n");
   
   for (i = 0; i < 256; i++)
     modifierPressed[i] = False;
@@ -126,22 +189,89 @@ Bool CreateXWindow(void)
   
   if(restingy == -1)
   {
-    restingy = si.framebufferHeight -2;
-    restingx = si.framebufferWidth -2;
+    restingy = si.framebufferHeight -2 - mac_mode;
+    restingx = si.framebufferWidth -2 + mac_mode;
   }
 
+#ifdef HAVE_XINERAMA
+ {
+   int x,y;
+   if(XineramaQueryExtension(dpy,&x,&y) &&
+      XineramaIsActive(dpy))
+   {
+     int pos,e;
+     int num_heads;
+     int bestpos=0;
+     int besthead=0;
+     XineramaScreenInfo *heads;
+
+     if(heads=XineramaQueryScreens(dpy, &num_heads))
+     {
+       /* Loop over all heads and find whatever head
+	* corresponds best with the edge the user has
+	* selected. If the user selects "north", the
+	* bigger screen will normally be selected.
+	*
+	* This is actually kind of stupid, it should really
+	* use the biggest screen for off-screen stuff.
+	*/
+       
+       for(e=0;e<num_heads;e++)
+       {
+	 switch(edge)
+	 {
+	   case EDGE_EAST:
+	     pos=heads[e].x_org + heads[e].width + (heads[e].height>>8);
+	     break;
+	   case EDGE_WEST:
+	     pos=1000-heads[e].x_org + (heads[e].height>>8);
+	     break;
+	   case EDGE_SOUTH:
+	     pos=heads[e].y_org + heads[e].height + (heads[e].width>>8);
+	     break;
+	   case EDGE_NORTH:
+	     pos=1000-heads[e].y_org + (heads[e].width>>8);
+	     break;
+	 }
+	 fprintf(stderr,"screen[%d] pos=%d\n",e,pos);
+
+	 if(pos > bestpos)
+	 {
+	   bestpos=pos;
+	   besthead=e;
+	 }
+       }
+
+
+       printf("Xinerama detected, x2vnc will use screen %d.\n",
+	      besthead+1);
+
+       x_offset = heads[besthead].x_org;
+       y_offset = heads[besthead].y_org;
+       displayWidth = heads[besthead].width;
+       displayHeight = heads[besthead].height;
+#if 0
+       fprintf(stderr,"[%d,%d-%d,%d]\n",x_offset,y_offset,
+	       displayWidth,displayHeight);
+#endif
+     }
+     XFree(heads);
+   }
+ }
+#endif
+
   topLevelWidth=edge_width;
   topLevelHeight=edge_width;
-  wmHints.x=0;
-  wmHints.y=0;
+  wmHints.x=x_offset;
+  wmHints.y=y_offset;
   
   switch(edge)
   {
-    case EDGE_EAST: wmHints.x=displayWidth-edge_width;
+    case EDGE_EAST: wmHints.x=displayWidth-edge_width+x_offset;
     case EDGE_WEST: topLevelHeight=displayHeight;
       break;
       
-    case EDGE_SOUTH: wmHints.y=displayHeight-edge_width;
+    case EDGE_SOUTH: wmHints.y=displayHeight-edge_width+y_offset;
     case EDGE_NORTH: topLevelWidth=displayWidth;
       break;
   }
@@ -247,12 +377,17 @@ Bool CreateXWindow(void)
   }
   
   nullPixmap = XCreatePixmap(dpy, DefaultRootWindow(dpy), 1, 1, 1);
+#if 0
+  grabCursor = 0;
+#else
   grabCursor = 
     XCreatePixmapCursor(dpy, nullPixmap, nullPixmap,
 			&dummyColor, &dummyColor, 0, 0);
+#endif
   
   if(grabkeysym)
   {
+    xerrorhandler prev_err_handler=XSetErrorHandler(warn_about_hotkey);
     grabkey=XKeysymToKeycode(dpy, grabkeysym);
 #ifdef DEBUG
     fprintf(stderr,"Grabbing key = %d, modifiers = %x\n",grabkey,grabmod);
@@ -261,15 +396,24 @@ Bool CreateXWindow(void)
 	     DefaultRootWindow(dpy),
 	     1, GrabModeSync, GrabModeSync);
 
-    /* Allow numlock */
-    XGrabKey(dpy, grabkey, grabmod | Mod2Mask,
-	     DefaultRootWindow(dpy),
-	     1, GrabModeSync, GrabModeSync);
+    XSync(dpy, 0);
+
+    if(grabkeysym)
+    {
+      /* Allow numlock */
+      XGrabKey(dpy, grabkey, grabmod | Mod2Mask,
+	       DefaultRootWindow(dpy),
+	       1, GrabModeSync, GrabModeSync);
+    }
+    XSync(dpy, 0);
+
+    XSetErrorHandler(prev_err_handler);
   }
 
   /* hide the cursor, so user won't get confused
      which keyboard has control */
   SendPointerEvent(restingx,restingy,0);
+
   
   return True;
 }
@@ -287,6 +431,46 @@ ShutdownX()
 }
 
 
+/*
+ * check_idle()
+ * Check how long the user has been idle using
+ * various methods.... (in ms)
+ */
+
+Time check_idle(void)
+{
+#ifdef HAVE_XIDLE
+  if(server_has_xidle)
+  {
+    Time ret = 0;
+    XGetIdleTime(dpy, &ret);
+    return ret;
+  }
+#endif
+
+#if HAVE_MIT_SCREENSAVER
+  if(server_has_mitscreensaver)
+  {
+    static XScreenSaverInfo* info = 0; 
+  
+    if (!info) info = XScreenSaverAllocInfo();
+    
+    XScreenSaverQueryInfo(dpy, DefaultRootWindow(dpy), info);
+    return info->idle;
+  }
+#endif
+
+  return 0;
+}
+
+void WiggleMouse(void)
+{
+  int tmpy=restingy > si.framebufferHeight /2 ? restingy -1 : restingy + 1;
+  SendPointerEvent(restingx,tmpy,0);
+  SendPointerEvent(restingx,restingy,0);
+}
+
+
 /*
  * HandleXEvents.
  */
@@ -295,11 +479,28 @@ Bool HandleXEvents(void)
 {
   XEvent ev;
 
-  if(grab_timeout_delay && grabbed && time(0) > grab_timeout)
+  if(grabbed)
   {
-    ungrabit(-1, -1, DefaultRootWindow(dpy));
+    if(grab_timeout_delay && time(0) > grab_timeout)
+      ungrabit(-1, -1, DefaultRootWindow(dpy));
+  }else{ /* grabbed */
+    int remote_idle = time(0) - last_event_time;
+    if(remote_idle > no_wakeup_delay)
+      remote_is_locked=1;
+#ifdef DEBUG
+    fprintf(stderr,"IDLE=%d remote_idle=%d\n",
+	    check_idle(),remote_idle);
+#endif
+    if(noblank &&
+       remote_idle > 45 &&
+       !remote_is_locked &&
+       check_idle() < 30000)
+    {
+      WiggleMouse();
+    }
   }
   
+  
   /* presumably XCheckMaskEvent is more efficient than XCheckIfEvent -GRM */
     while (XCheckIfEvent(dpy, &ev, AllXEventsPredicate, NULL))
 /*  while (XCheckMaskEvent(dpy, ~0, &ev)) */
@@ -380,7 +581,9 @@ static void grabit(int x, int y, int state)
 
   if(x > -1 && y > -1)
   {
-    XWarpPointer(dpy,None, DefaultRootWindow(dpy),0,0,0,0, x, y);
+    XWarpPointer(dpy,None, DefaultRootWindow(dpy),0,0,0,0,
+		 x_offset + x,
+		 y_offset + y);
     sendpointerevent(x, y, (state & 0x1f00) >> 8);
   }
 
@@ -404,7 +607,7 @@ static void ungrabit(int x, int y, Window warpWindow)
   int i;
   SendPointerEvent(restingx,restingy,0);
   if(x > -1 && y > -1 )
-    XWarpPointer(dpy,None, warpWindow, 0,0,0,0, x, y);
+    XWarpPointer(dpy,None, warpWindow, 0,0,0,0, x_offset + x, y_offset + y);
   XUngrabKeyboard(dpy, CurrentTime);
   XUngrabPointer(dpy, CurrentTime);
   mouseOnScreen = warpWindow == DefaultRootWindow(dpy);
@@ -557,8 +760,8 @@ static Bool HandleTopLevelEvent(XEvent *ev)
     case EnterNotify:
       if(!grabbed && ev->xcrossing.mode==NotifyNormal)
       {
-	grabit(enter_translate(EW,displayWidth ,ev->xcrossing.x_root),
-	       enter_translate(NS,displayHeight,ev->xcrossing.y_root),
+	grabit(enter_translate(EW,displayWidth ,XROOT(ev->xcrossing)),
+	       enter_translate(NS,displayHeight,YROOT(ev->xcrossing)),
 	       ev->xcrossing.state);
       }
       return 1;
@@ -575,12 +778,13 @@ static Bool HandleTopLevelEvent(XEvent *ev)
 
 #if 0
 	fprintf(stderr,"{ %d, %d } -> { %d, %d }\n",
-		ev->xmotion.x_root,
-		ev->xmotion.y_root,
-		SCALEX(ev->xmotion.x_root),
-		SCALEY(ev->xmotion.y_root));
+		XROOT(ev->xmotion),
+		YROOT(ev->xmotion),
+		SCALEX(XROOT(ev->xmotion)),
+		SCALEY(YROOT(ev->xmotion));
 #endif
-	i=sendpointerevent(ev->xmotion.x_root, ev->xmotion.y_root,
+	i=sendpointerevent(XROOT(ev->xmotion),
+			   YROOT(ev->xmotion),
 			   (ev->xmotion.state & 0x1f00) >> 8);
 
 	if(ev->xmotion.state & 0x1f00) return 1;
@@ -596,8 +800,8 @@ static Bool HandleTopLevelEvent(XEvent *ev)
          */
         if (ev->xmotion.same_screen)
 	{
-          x = ev->xmotion.x_root;
-          y = ev->xmotion.y_root;
+          x = XROOT(ev->xmotion);
+          y = YROOT(ev->xmotion);
           warpWindow = DefaultRootWindow(dpy);
 	  switch(edge)
 	  {
@@ -636,21 +840,21 @@ static Bool HandleTopLevelEvent(XEvent *ev)
              * - GRM
              */
           warpWindow = ev->xmotion.root;
-          x = ev->xmotion.x_root;
-          y = ev->xmotion.y_root;
+          x = XROOT(ev->xmotion);
+          y = YROOT(ev->xmotion);
 	  switch(edge)
 	  {
 	    case EDGE_NORTH:
-              d=ev->xmotion.y_root < displayHeight / 2;
+              d=YROOT(ev->xmotion) < displayHeight / 2;
               break;
 	    case EDGE_SOUTH:
-              d=ev->xmotion.y_root > displayHeight / 2;
+              d=YROOT(ev->xmotion) > displayHeight / 2;
               break;
 	    case EDGE_EAST:
-              d=ev->xmotion.x_root > displayWidth / 2;
+              d=XROOT(ev->xmotion) > displayWidth / 2;
               break;
 	    case EDGE_WEST:
-              d=ev->xmotion.x_root < displayWidth / 2;
+              d=XROOT(ev->xmotion) < displayWidth / 2;
               break;
 	  }
           if (d) {
@@ -706,8 +910,8 @@ static Bool HandleTopLevelEvent(XEvent *ev)
 		      ~(1 << (ev->xbutton.button - 1)));
       }
       
-      return sendpointerevent(ev->xbutton.x_root,
-			      ev->xbutton.y_root,
+      return sendpointerevent(XROOT(ev->xbutton),
+			      YROOT(ev->xbutton),
 			      buttonMask);
 	
     case KeyPress:
@@ -725,8 +929,8 @@ static Bool HandleTopLevelEvent(XEvent *ev)
 	 (ev->xkey.state == grabmod ||
 	  (ev->xkey.state & ~ Mod2Mask ) == grabmod ))
       {
-	saved_remote_xpos=ev->xkey.x_root;
-	saved_remote_ypos=ev->xkey.y_root;
+	saved_remote_xpos=XROOT(ev->xkey);
+	saved_remote_ypos=YROOT(ev->xkey);
 	ungrabit(saved_xpos,saved_ypos,DefaultRootWindow(dpy));
 	return 1;
       }
@@ -749,6 +953,9 @@ static Bool HandleTopLevelEvent(XEvent *ev)
       }
 /*      fprintf(stderr,"-------- %x (%c) name=%s\n",ks,ks,keyname); */
 
+      /* We assume that typing means an unlocked display */
+      remote_is_locked=0;
+
       return SendKeyEvent(ks, (ev->type == KeyPress));
     }
       
@@ -854,8 +1061,8 @@ static Bool HandleRootEvent(XEvent *ev)
 #endif
       if(ev->xkey.keycode)
       {
-	saved_xpos=ev->xkey.x_root;
-	saved_ypos=ev->xkey.y_root;
+	saved_xpos=XROOT(ev->xkey);
+	saved_ypos=YROOT(ev->xkey);
 	grabit(saved_remote_xpos, saved_remote_ypos, 0);
       }
       break;
@@ -887,11 +1094,11 @@ static Bool HandleRootEvent(XEvent *ev)
 	   * 
 	   * - GRM
 	   */
-	  x = ev->xcrossing.x_root;
-	  y = ev->xcrossing.x_root;
+	  x = XROOT(ev->xcrossing);
+	  y = XROOT(ev->xcrossing);
 	  if (!nowOnScreen) {
-	    x = enter_translate(EW,displayWidth,ev->xcrossing.x_root);
-	    y = enter_translate(NS,displayHeight,ev->xcrossing.y_root);
+	    x = enter_translate(EW,displayWidth,XROOT(ev->xcrossing));
+	    y = enter_translate(NS,displayHeight,YROOT(ev->xcrossing));
 	  }
 	  switch(edge)
 	  {
@@ -920,8 +1127,8 @@ static Bool HandleRootEvent(XEvent *ev)
        */
       if(grab && ev->xcrossing.mode == NotifyNormal)
       {
-	grabit(enter_translate(EW,displayWidth ,ev->xcrossing.x_root),
-	       enter_translate(NS,displayHeight,ev->xcrossing.y_root),
+	grabit(enter_translate(EW,displayWidth ,XROOT(ev->xcrossing)),
+	       enter_translate(NS,displayHeight,YROOT(ev->xcrossing)),
 	       ev->xcrossing.state);
       }
       break;
